#! /usr/bin/env python

import random
import numpy as np

import sys

fname = sys.argv[1]
dat = np.genfromtxt(fname,names=True)
p = np.random.permutation(len(dat))
dat_out  = dat[p]
fname_out = sys.argv[2]
np.savetxt(fname_out,dat_out,header='# ' + ' '.join(dat.dtype.names))



#! /usr/bin/env python

import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import scipy
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import functools
import itertools


parser = argparse.ArgumentParser()
parser.add_argument("--fname-out", default="output-grid.dat", help="Name of dat file")
parser.add_argument("--npts",default=100,type=int)
parser.add_argument("--random-parameter", action='append',help="These parameters are specified at random over the entire range, uncorrelated with the grid used for other parameters.  Use for variables which correlate weakly with others; helps with random exploration")
parser.add_argument("--random-parameter-range", action='append', type=str,help="Add a range (pass as a string evaluating to a python 2-element list): --parameter-range '[0.,1000.]'   MUST specify ALL parameter ranges (min and max) in order if used.  ")
opts=  parser.parse_args()

npts = opts.npts

# Extract parameter names
coord_names = opts.random_parameter # Used  in fit

param_ranges = list(map(eval, opts.random_parameter_range))  # do not randomize mass or distance parameters, not being rescaled

#if opts.parameter_nofit:
#    coord_names = coord_names + opts.parameter_nofit
if coord_names is None:
    sys.exit(0)


# Load data, keep parameter names
X= np.zeros((npts, 2+len(coord_names)))
for indx in np.arange(len(coord_names)):
        X[:,2+indx] = np.random.uniform(param_ranges[indx][0], param_ranges[indx][1],size=npts)

np.savetxt(opts.fname_out, X,header="lnL sigma_lnL " + " ".join(coord_names))



#! /usr/bin/env python
#
#
# INTENT
#    - retrieve external grid(s), deployed in target format, and merged
#    - used to create cross-run dependencies in iterative structure

# SPECIFICATION: argument file

# * json file
#   label
#   method    : file
#     label
#     command : some methods require commands
#     arguments  : command line arguments.  Note this will not include a target name, which must be allowed as the next argument
#   convert   : None, or block
#     label
#     command
#     arguments
#
# Example
#   {
    #   "method" : "native",
    #   "source":  rundir   # will take latest grid directly, verbatim
    #   "n_max": 1000   # cap on number of points taken
    # }
    # }
# target methods
#    - copy (copy existing file)
#    - copy_latest (copy last item matching pattern, using numerical sort)



# **FETCH SPECIFICATION**


# * json file.  Default is ONE JSON PER EXECUTABLE.
#   label
#   method    : file
#     label
#     command : some methods require commands
#     arguments  : command line arguments
#   convert   : None, or block
#     label
#     command
#     arguments
   

from RIFT.misc.samples_utils import add_field,extract_combination_from_LI
import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import functools
import itertools
import os

import json
import shutil, glob,re


def retrieve_native(sourcedir,outfile,n_max=None,base_pattern="overlap-grid-*.xml.gz",verbose=True):
    """
    retrieve_native(sourcedir,outfile)

    sourcedir : source directory.  Looks for last file of form "output-grid-N.xml.gz
    outfile : target file name, assume it is full path
    n_max :
    """
    if verbose:
        print("Checking ", sourcedir, " for ", base_pattern)
    # Identify the correct source file in the directory
    fnames = glob.glob(sourcedir+"/"+base_pattern)  # give flexibility to naming/reuse of this code
    fnames.sort(key=lambda f:int(re.sub('\D', '', f))) # trick from https://pretagteam.com/question/how-to-sort-files-by-number-in-filename-of-a-directory-duplicate
    # if verbose:
    #     print(fnames)
    fname_to_use = fnames[-1]

    # If n_max is not None, load in the file, truncate its size
    if n_max is None:
        if verbose:
            print(" Transferring ", fname_to_use, " -> ", outfile)
        shutil.copyfile(fname_to_use, outfile)
    elif n_max > 0:
        import random
        import RIFT.lalsimutils as lalsimutils
        # Load in grid
        P_list = lalsimutils.xml_to_ChooseWaveformParams_array(fname_to_use)
        # select points randomly!
        P_list_reduced = random.sample(P_list, int(n_max))
        lalsimutils.ChooseWaveformParams_array_to_xml(P_list, outfile)
    else:
        print(" Invalid fetch size ", n_max)
        import sys; sys.exit(99)
    return None


parser = argparse.ArgumentParser()
parser.add_argument("--input-json",type=str,default="fetch.json",help="input file")
parser.add_argument("--inj-file-out",type=str,default="merged_grid",help="output file")
opts=  parser.parse_args()


rundir = os.getcwd()

config=None
with open(opts.input_json,'r') as f:
    config = json.load(f)

method = config['method']
if method =='native':
    retrieve_native(config['source'],opts.inj_file_out)



#! /usr/bin/env python
#
# USAGE
#  python util_TestXMLParameterStorage.py ; ligolw_print -t sim_inspiral -c alpha5 output.xml.gz
import RIFT.lalsimutils as lalsimutils
import numpy as np
import lal

import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--output-file", default="output")
opts = parser.parse_args()

P_list_in =[]
for i in np.arange(5):
    P_list_in.append( lalsimutils.ChooseWaveformParams(m1=lal.MSUN_SI*(i+1)))
    P_list_in[-1].lambda1 = 3*i+1
    P_list_in[-1].lambda2 = 3*i+2
    P_list_in[-1].approx = 3*i+2
    P_list_in[-1].print_params()

lalsimutils.ChooseWaveformParams_array_to_xml(P_list_in, fname=opts.output_file)

print(" ------ ")
P_list_out = lalsimutils.xml_to_ChooseWaveformParams_array(opts.output_file+".xml.gz")

for P in P_list_out:
    P.print_params()



#! /usr/bin/env python
import numpy as np
import RIFT.lalsimutils as lalsimutils

import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--fname",help="filename of *.dat file [standard ILE output]")
parser.add_argument("--fname-reference",help="xml also")
parser.add_argument("--parameter",action='append',help="list of parameter names")
opts=  parser.parse_args()

P_list = lalsimutils.xml_to_ChooseWaveformParams_array(opts.fname)
P_ref_list= lalsimutils.xml_to_ChooseWaveformParams_array(opts.fname_reference)

# Find the mean value of the reference point
parameters_to_center = ['mc', 'eta']   # could also add 'chieff_aligned', in cases with significant spin
if opts.parameter:
   parameters_to_center =opts.parameter
dat = np.zeros((len(P_list), len(parameters_to_center)))
for indx in np.arange(len(P_list)):
   for pIndex in np.arange(len(parameters_to_center)):
      param = parameters_to_center[pIndex]
      val = P_list[indx].extract_param(param)
      dat[indx,pIndex]=val
dat_ref = np.zeros((len(P_ref_list), len(parameters_to_center)))
for indx in np.arange(len(P_ref_list)):
   for pIndex in np.arange(len(parameters_to_center)):
        param = parameters_to_center[pIndex]
        val = P_ref_list[indx].extract_param(param)
        dat_ref[indx,pIndex]=val

P_out = []
dx = np.mean(dat,axis=0) - np.mean(dat_ref,axis=0)
dat_out = []
for indx in np.arange(len(P_list)):
   bInclude = True
   vec = np.zeros(len(parameters_to_center))
   for pIndex in np.arange(len(parameters_to_center)):
      param =  parameters_to_center[pIndex]
      vec[pIndex] = valNew = dat[indx,pIndex]-dx[pIndex]
      if param == 'eta':
         if valNew > 0.25 or valNew < 0.001:
            bInclude=False
            continue
      if param == 's1z' or param == 's2z':
         if valNew > 1 or valNew < -1:
            bInclude=False
            continue
      P_list[indx].assign_param(param, valNew)
   P_list[indx].tref = float(P_list[indx].tref)
   P_out.append(P_list[indx])
   dat_out.append(vec)

print(" Recentering report: last two should be equal : ", np.mean(dat,axis=0),np.mean(dat_ref,axis=0),np.mean(dat_out,axis=0))

lalsimutils.ChooseWaveformParams_array_to_xml(P_out, "shifted.xml.gz")



# Dan Wysocki

import numpy as np
import numpy

def histogram(samples, n_bins, xpy=numpy,weights=None):
    n_samples = samples.size

    # Compute the histogram counts.
    indices = xpy.trunc(samples * n_bins).astype(np.int32)
    if isinstance(weights,type(None)):
        wts  =xpy.broadcast_to(
            xpy.asarray([float(n_bins)/n_samples]),
            (n_samples,)
            )
    else:
        wts=weights
    histogram_counts = xpy.bincount(
        indices, minlength=n_bins,
        weights=wts
    )
    return histogram_counts[:n_bins]  # force target length, we should never have points in top bin if it occurs : scaled to [0,1)



def interp(x, xp, fp, left=None, right=None, period=None, xpy=numpy):
    """
    One-dimensional linear interpolation.

    Returns the one-dimensional piecewise linear interpolant to a function
    with given discrete data points (`xp`, `fp`), evaluated at `x`.

    Parameters
    ----------
    x : array_like
        The x-coordinates at which to evaluate the interpolated values.

    xp : 1-D sequence of floats
        The x-coordinates of the data points, must be increasing if argument
        `period` is not specified. Otherwise, `xp` is internally sorted after
        normalizing the periodic boundaries with ``xp = xp % period``.

    fp : 1-D sequence of float or complex
        The y-coordinates of the data points, same length as `xp`.

    left : optional float or complex corresponding to fp
        Value to return for `x < xp[0]`, default is `fp[0]`.

    right : optional float or complex corresponding to fp
        Value to return for `x > xp[-1]`, default is `fp[-1]`.

    period : None or float, optional
        A period for the x-coordinates. This parameter allows the proper
        interpolation of angular x-coordinates. Parameters `left` and `right`
        are ignored if `period` is specified.

        .. versionadded:: 1.10.0

    Returns
    -------
    y : float or complex (corresponding to fp) or ndarray
        The interpolated values, same shape as `x`.

    Raises
    ------
    ValueError
        If `xp` and `fp` have different length
        If `xp` or `fp` are not 1-D sequences
        If `period == 0`

    Notes
    -----
    Does not check that the x-coordinate sequence `xp` is increasing.
    If `xp` is not increasing, the results are nonsense.
    A simple check for increasing is::

        np.all(np.diff(xp) > 0)
    """
    # TODO: Implement periodic interpolation if needed.
    if period is not None:
        raise NotImplementedError("Periodic interpolation not yet implemented.")

    # Check shapes.
    input_shape = xpy.shape(xp)
    output_shape = xpy.shape(x)

    # Validate shapes.
    if len(input_shape) != 1:
        raise ValueError("`xp` is not a 1-D sequence")
    if input_shape != xpy.shape(fp):
        raise ValueError("`xp` and `fp` have different lengths.")

    # Pull out number of samples, now that we know input is 1-D.
    n_samples, = input_shape

    # Fill in default values for left and right if not given
    if left is None:
        left = fp[0]
    if right is None:
        right = fp[-1]

    # Initialize output array.
    y = xpy.empty_like(x, dtype=fp.dtype)

    # Compute (f(x1)-f(x0)) / (x1-x0)
    slopes = xpy.diff(fp) / xpy.diff(xp)

    # Get indices of different regions.
    i_below = x <= xp[0]
    i_above = x >= xp[-1]
    i_inside = ~(i_below | i_above)

    # Process points beyond edges.
    y[i_below] = left
    y[i_above] = right

    # Process interior.  First get indices corresponding to bins.
    ## TODO: Vectorize.  Tricky without `np.searchsorted` or `np.digitize`.
    x_inside = x[i_inside]
    x_bin_indices = xpy.empty_like(x_inside, dtype=int)
    for i in numpy.ndindex(*x_inside.shape):
        x_bin_indices[i] = xpy.argmax(
            (xp[:-1] < x_inside[i]) & (x_inside[i] <= xp[1:])
        )

    y[i_inside] = (
        slopes[x_bin_indices]*(x_inside-xp[x_bin_indices]) + fp[x_bin_indices]
    )

    return y



"""
Gaussian Mixture Models.

This implementation corresponds to frequentist (non-Bayesian) formulation
of Gaussian Mixture Models.

Adapted from sklearn, added ability to specify weights.
"""

# Author: Ron Weiss <ronweiss@gmail.com>
#         Fabian Pedregosa <fabian.pedregosa@inria.fr>
#         Bertrand Thirion <bertrand.thirion@inria.fr>

import warnings
import numpy as np
from scipy import linalg
from time import time

from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state, check_array
#from sklearn.utils.extmath import logsumexp
from scipy.special import logsumexp  # error in scipy docs
from sklearn.utils.validation import check_is_fitted
from sklearn import cluster

from six.moves import zip

EPS = np.finfo(float).eps


def log_multivariate_normal_density(X, means, covars, covariance_type='diag'):
    """Compute the log probability under a multivariate Gaussian distribution.

    Parameters
    ----------
    X : array_like, shape (n_samples, n_features)
        List of n_features-dimensional data points.  Each row corresponds to a
        single data point.

    means : array_like, shape (n_components, n_features)
        List of n_features-dimensional mean vectors for n_components Gaussians.
        Each row corresponds to a single mean vector.

    covars : array_like
        List of n_components covariance parameters for each Gaussian. The shape
        depends on `covariance_type`:
            (n_components, n_features)      if 'spherical',
            (n_features, n_features)    if 'tied',
            (n_components, n_features)    if 'diag',
            (n_components, n_features, n_features) if 'full'

    covariance_type : string
        Type of the covariance parameters.  Must be one of
        'spherical', 'tied', 'diag', 'full'.  Defaults to 'diag'.

    Returns
    -------
    lpr : array_like, shape (n_samples, n_components)
        Array containing the log probabilities of each data point in
        X under each of the n_components multivariate Gaussian distributions.
    """
    log_multivariate_normal_density_dict = {
        'spherical': _log_multivariate_normal_density_spherical,
        'tied': _log_multivariate_normal_density_tied,
        'diag': _log_multivariate_normal_density_diag,
        'full': _log_multivariate_normal_density_full}
    return log_multivariate_normal_density_dict[covariance_type](
        X, means, covars)


def sample_gaussian(mean, covar, covariance_type='diag', n_samples=1,
                    random_state=None):
    """Generate random samples from a Gaussian distribution.

    Parameters
    ----------
    mean : array_like, shape (n_features,)
        Mean of the distribution.

    covar : array_like, optional
        Covariance of the distribution. The shape depends on `covariance_type`:
            scalar if 'spherical',
            (n_features) if 'diag',
            (n_features, n_features)  if 'tied', or 'full'

    covariance_type : string, optional
        Type of the covariance parameters.  Must be one of
        'spherical', 'tied', 'diag', 'full'.  Defaults to 'diag'.

    n_samples : int, optional
        Number of samples to generate. Defaults to 1.

    Returns
    -------
    X : array, shape (n_features, n_samples)
        Randomly generated sample
    """
    rng = check_random_state(random_state)
    n_dim = len(mean)
    rand = rng.randn(n_dim, n_samples)
    if n_samples == 1:
        rand.shape = (n_dim,)

    if covariance_type == 'spherical':
        rand *= np.sqrt(covar)
    elif covariance_type == 'diag':
        rand = np.dot(np.diag(np.sqrt(covar)), rand)
    else:
        s, U = linalg.eigh(covar)
        s.clip(0, out=s)        # get rid of tiny negatives
        np.sqrt(s, out=s)
        U *= s
        rand = np.dot(U, rand)

    return (rand.T + mean).T


class WeightedGMM(BaseEstimator):
    """Gaussian Mixture Model with weighted samples

    Representation of a Gaussian mixture model probability distribution.
    This class allows for easy evaluation of, sampling from, and
    maximum-likelihood estimation of the parameters of a GMM distribution.

    Initializes parameters such that every mixture component has zero
    mean and identity covariance.

    Read more in the :ref:`User Guide <gmm>`.

    Parameters
    ----------
    n_components : int, optional
        Number of mixture components. Defaults to 1.

    covariance_type : string, optional
        String describing the type of covariance parameters to
        use.  Must be one of 'spherical', 'tied', 'diag', 'full'.
        Defaults to 'diag'.

    random_state: RandomState or an int seed (None by default)
        A random number generator instance

    min_covar : float, optional
        Floor on the diagonal of the covariance matrix to prevent
        overfitting.  Defaults to 1e-3.

    tol : float, optional
        Convergence threshold. EM iterations will stop when average
        gain in log-likelihood is below this threshold.  Defaults to 1e-3.

    n_iter : int, optional
        Number of EM iterations to perform.

    n_init : int, optional
        Number of initializations to perform. the best results is kept

    params : string, optional
        Controls which parameters are updated in the training
        process.  Can contain any combination of 'w' for weights,
        'm' for means, and 'c' for covars.  Defaults to 'wmc'.

    init_params : string, optional
        Controls which parameters are updated in the initialization
        process.  Can contain any combination of 'w' for weights,
        'm' for means, and 'c' for covars.  Defaults to 'wmc'.

    verbose : int, default: 0
        Enable verbose output. If 1 then it always prints the current
        initialization and iteration step. If greater than 1 then
        it prints additionally the change and time needed for each step.

    Attributes
    ----------
    weights_ : array, shape (`n_components`,)
        This attribute stores the mixing weights for each mixture component.

    means_ : array, shape (`n_components`, `n_features`)
        Mean parameters for each mixture component.

    covars_ : array
        Covariance parameters for each mixture component.  The shape
        depends on `covariance_type`::

            (n_components, n_features)             if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'

    converged_ : bool
        True when convergence was reached in fit(), False otherwise.

    See Also
    --------

    DPGMM : Infinite gaussian mixture model, using the dirichlet
        process, fit with a variational algorithm


    VBGMM : Finite gaussian mixture model fit with a variational
        algorithm, better for situations where there might be too little
        data to get a good estimate of the covariance matrix.

    Examples
    --------

    >>> import numpy as np
    >>> from sklearn import mixture
    >>> np.random.seed(1)
    >>> g = mixture.GMM(n_components=2)
    >>> # Generate random observations with two modes centered on 0
    >>> # and 10 to use for training.
    >>> obs = np.concatenate((np.random.randn(100, 1),
    ...                       10 + np.random.randn(300, 1)))
    >>> g.fit(obs) # doctest: +NORMALIZE_WHITESPACE
    GMM(covariance_type='diag', init_params='wmc', min_covar=0.001,
            n_components=2, n_init=1, n_iter=100, params='wmc',
            random_state=None, thresh=None, tol=0.001, verbose=0)
    >>> np.round(g.weights_, 2)
    array([ 0.75,  0.25])
    >>> np.round(g.means_, 2)
    array([[ 10.05],
           [  0.06]])
    >>> np.round(g.covars_, 2) #doctest: +SKIP
    array([[[ 1.02]],
           [[ 0.96]]])
    >>> g.predict([[0], [2], [9], [10]]) #doctest: +ELLIPSIS
    array([1, 1, 0, 0]...)
    >>> np.round(g.score([[0], [2], [9], [10]]), 2)
    array([-2.19, -4.58, -1.75, -1.21])
    >>> # Refit the model on new data (initial parameters remain the
    >>> # same), this time with an even split between the two modes.
    >>> g.fit(20 * [[0]] +  20 * [[10]]) # doctest: +NORMALIZE_WHITESPACE
    GMM(covariance_type='diag', init_params='wmc', min_covar=0.001,
            n_components=2, n_init=1, n_iter=100, params='wmc',
            random_state=None, thresh=None, tol=0.001, verbose=0)
    >>> np.round(g.weights_, 2)
    array([ 0.5,  0.5])

    """

    def __init__(self, n_components=1, covariance_type='diag',
                 random_state=None, thresh=None, tol=1e-3, min_covar=1e-3,
                 n_iter=100, n_init=1, params='wmc', init_params='wmc',
                 verbose=0):
        if thresh is not None:
            warnings.warn("'thresh' has been replaced by 'tol' in 0.16 "
                          " and will be removed in 0.18.",
                          DeprecationWarning)
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.thresh = thresh
        self.tol = tol
        self.min_covar = min_covar
        self.random_state = random_state
        self.n_iter = n_iter
        self.n_init = n_init
        self.params = params
        self.init_params = init_params
        self.verbose = verbose

        if covariance_type not in ['spherical', 'tied', 'diag', 'full']:
            raise ValueError('Invalid value for covariance_type: %s' %
                             covariance_type)

        if n_init < 1:
            raise ValueError('GMM estimation requires at least one run')

        self.weights_ = np.ones(self.n_components) / self.n_components

        # flag to indicate exit status of fit() method: converged (True) or
        # n_iter reached (False)
        self.converged_ = False

    def _get_covars(self):
        """Covariance parameters for each mixture component.

        The shape depends on ``cvtype``::

            (n_states, n_features)                if 'spherical',
            (n_features, n_features)              if 'tied',
            (n_states, n_features)                if 'diag',
            (n_states, n_features, n_features)    if 'full'

        """
        if self.covariance_type == 'full':
            return self.covars_
        elif self.covariance_type == 'diag':
            return [np.diag(cov) for cov in self.covars_]
        elif self.covariance_type == 'tied':
            return [self.covars_] * self.n_components
        elif self.covariance_type == 'spherical':
            return [np.diag(cov) for cov in self.covars_]

    def _set_covars(self, covars):
        """Provide values for covariance"""
        covars = np.asarray(covars)
        _validate_covars(covars, self.covariance_type, self.n_components)
        self.covars_ = covars

    def score_samples(self, X, w=None):
        """Return the per-sample likelihood of the data under the model.

        Compute the log probability of X under the model and
        return the posterior distribution (responsibilities) of each
        mixture component for each element of X.

        Parameters
        ----------
        X: array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        w : array-like, shape = [n_samples] (optional)
            Sample weights

        Returns
        -------
        logprob : array_like, shape (n_samples,)
            Log probabilities of each data point in X.

        responsibilities : array_like, shape (n_samples, n_components)
            Posterior probabilities of each mixture component for each
            observation
        """
        check_is_fitted(self, 'means_')

        X = check_array(X)
        if X.ndim == 1:
            X = X[:, np.newaxis]
        if X.size == 0:
            return np.array([]), np.empty((0, self.n_components))
        if X.shape[1] != self.means_.shape[1]:
            raise ValueError('The shape of X  is not compatible with self')

        lpr = (log_multivariate_normal_density(X, self.means_, self.covars_,
                                               self.covariance_type) +
               np.log(self.weights_))
        logprob = logsumexp(lpr, axis=1)
        responsibilities = np.exp(lpr - logprob[:, np.newaxis])
        # Apply sample weighting
        if w is not None:
            responsibilities *= w

        return logprob, responsibilities

    def score(self, X, w=None, y=None):
        """Compute the log probability under the model.

        Parameters
        ----------
        X : array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        Returns
        -------
        logprob : array_like, shape (n_samples,)
            Log probabilities of each data point in X
        """
        logprob, _ = self.score_samples(X, w)
        return logprob

    def predict(self, X):
        """Predict label for data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]

        Returns
        -------
        C : array, shape = (n_samples,) component memberships
        """
        logprob, responsibilities = self.score_samples(X)
        return responsibilities.argmax(axis=1)

    def predict_proba(self, X, w=None):
        """Predict posterior probability of data under each Gaussian
        in the model.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]

        Returns
        -------
        responsibilities : array-like, shape = (n_samples, n_components)
            Returns the probability of the sample for each Gaussian
            (state) in the model.
        """
        logprob, responsibilities = self.score_samples(X, w)
        return responsibilities

    def sample(self, n_samples=1, random_state=None):
        """Generate random samples from the model.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples to generate. Defaults to 1.

        Returns
        -------
        X : array_like, shape (n_samples, n_features)
            List of samples
        """
        check_is_fitted(self, 'means_')

        if random_state is None:
            random_state = self.random_state
        random_state = check_random_state(random_state)
        weight_cdf = np.cumsum(self.weights_)

        X = np.empty((n_samples, self.means_.shape[1]))
        rand = random_state.rand(n_samples)
        # decide which component to use for each sample
        comps = weight_cdf.searchsorted(rand)
        # for each component, generate all needed samples
        for comp in range(self.n_components):
            # occurrences of current component in X
            comp_in_X = (comp == comps)
            # number of those occurrences
            num_comp_in_X = comp_in_X.sum()
            if num_comp_in_X > 0:
                if self.covariance_type == 'tied':
                    cv = self.covars_
                elif self.covariance_type == 'spherical':
                    cv = self.covars_[comp][0]
                else:
                    cv = self.covars_[comp]
                X[comp_in_X] = sample_gaussian(
                    self.means_[comp], cv, self.covariance_type,
                    num_comp_in_X, random_state=random_state).T
        return X

    def fit_predict(self, X, w=None, y=None):
        """Fit and then predict labels for data.

        Warning: due to the final maximization step in the EM algorithm,
        with low iterations the prediction may not be 100% accurate

        .. versionadded:: 0.17
           *fit_predict* method in Gaussian Mixture Model.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        w : array-like, shape = [n_samples] (optional)
            Sample weights

        Returns
        -------
        C : array, shape = (n_samples,) component memberships
        """
        return self._fit(X, w, y).argmax(axis=1)

    def _fit(self, X, w=None, y=None, do_prediction=False):
        """Estimate model parameters with the EM algorithm.

        A initialization step is performed before entering the
        expectation-maximization (EM) algorithm. If you want to avoid
        this step, set the keyword argument init_params to the empty
        string '' when creating the GMM object. Likewise, if you would
        like just to do an initialization, set n_iter=0.

        Parameters
        ----------
        X : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        w : array-like, shape = [n_samples] (optional)
            Sample weights

        Returns
        -------
        responsibilities : array, shape (n_samples, n_components)
            Posterior probabilities of each mixture component for each
            observation.
        """

        # initialization step
        X = check_array(X, dtype=np.float64, ensure_min_samples=2,
                        estimator=self)
        if X.shape[0] < self.n_components:
            raise ValueError(
                'GMM estimation with %s components, but got only %s samples' %
                (self.n_components, X.shape[0]))

        max_log_prob = -np.infty

        if self.verbose > 0:
            print('Expectation-maximization algorithm started.')

        for init in range(self.n_init):
            if self.verbose > 0:
                print('Initialization ' + str(init + 1))
                start_init_time = time()

            if 'm' in self.init_params or not hasattr(self, 'means_'):
                self.means_ = cluster.KMeans(
                    n_clusters=self.n_components,
                    random_state=self.random_state).fit(X).cluster_centers_
                if self.verbose > 1:
                    print('\tMeans have been initialized.')

            if 'w' in self.init_params or not hasattr(self, 'weights_'):
                self.weights_ = np.tile(1.0 / self.n_components,
                                        self.n_components)
                if self.verbose > 1:
                    print('\tWeights have been initialized.')

            if 'c' in self.init_params or not hasattr(self, 'covars_'):
                cv = np.cov(X.T) + self.min_covar * np.eye(X.shape[1])
                if not cv.shape:
                    cv.shape = (1, 1)
                self.covars_ = \
                    distribute_covar_matrix_to_match_covariance_type(
                        cv, self.covariance_type, self.n_components)
                if self.verbose > 1:
                    print('\tCovariance matrices have been initialized.')

            # EM algorithms
            current_log_likelihood = None
            # reset self.converged_ to False
            self.converged_ = False

            # this line should be removed when 'thresh' is removed in v0.18
            tol = (self.tol if self.thresh is None
                   else self.thresh / float(X.shape[0]))

            for i in range(self.n_iter):
                if self.verbose > 0:
                    print('\tEM iteration ' + str(i + 1))
                    start_iter_time = time()
                prev_log_likelihood = current_log_likelihood
                # Expectation step
                log_likelihoods, responsibilities = self.score_samples(X, w)
                current_log_likelihood = log_likelihoods.mean()

                # Check for convergence.
                # (should compare to self.tol when deprecated 'thresh' is
                # removed in v0.18)
                if prev_log_likelihood is not None:
                    change = abs(current_log_likelihood - prev_log_likelihood)
                    if self.verbose > 1:
                        print('\t\tChange: ' + str(change))
                    if change < tol:
                        self.converged_ = True
                        if self.verbose > 0:
                            print('\t\tEM algorithm converged.')
                        break

                # Maximization step
                self._do_mstep(X, w, responsibilities, self.params,
                               self.min_covar)
                if self.verbose > 1:
                    print('\t\tEM iteration ' + str(i + 1) + ' took {0:.5f}s'.format(
                        time() - start_iter_time))

            # if the results are better, keep it
            if self.n_iter:
                if current_log_likelihood > max_log_prob:
                    max_log_prob = current_log_likelihood
                    best_params = {'weights': self.weights_,
                                   'means': self.means_,
                                   'covars': self.covars_}
                    if self.verbose > 1:
                        print('\tBetter parameters were found.')

            if self.verbose > 1:
                print('\tInitialization ' + str(init + 1) + ' took {0:.5f}s'.format(
                    time() - start_init_time))

        # check the existence of an init param that was not subject to
        # likelihood computation issue.
        if np.isneginf(max_log_prob) and self.n_iter:
            raise RuntimeError(
                "EM algorithm was never able to compute a valid likelihood " +
                "given initial parameters. Try different init parameters " +
                "(or increasing n_init) or check for degenerate data.")

        if self.n_iter:
            self.covars_ = best_params['covars']
            self.means_ = best_params['means']
            self.weights_ = best_params['weights']
        else:  # self.n_iter == 0 occurs when using GMM within HMM
            # Need to make sure that there are responsibilities to output
            # Output zeros because it was just a quick initialization
            responsibilities = np.zeros((X.shape[0], self.n_components))

        return responsibilities

    def fit(self, X, w=None, y=None):
        """Estimate model parameters with the EM algorithm.

        A initialization step is performed before entering the
        expectation-maximization (EM) algorithm. If you want to avoid
        this step, set the keyword argument init_params to the empty
        string '' when creating the GMM object. Likewise, if you would
        like just to do an initialization, set n_iter=0.

        Parameters
        ----------
        X : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        w : array-like, shape = [n_samples] (optional)
            Sample weights

        Returns
        -------
        self
        """
        self._fit(X, w, y)
        return self

    def _do_mstep(self, X, w, responsibilities, params, min_covar=0):
        """ Perform the Mstep of the EM algorithm and return the class weights
        """
        weights = responsibilities.sum(axis=0)
        weighted_X_sum = np.dot(responsibilities.T, X)
        inverse_weights = 1.0 / (weights[:, np.newaxis] + 10 * EPS)

        if 'm' in params:
            self.means_ = weighted_X_sum * inverse_weights
        if 'w' in params:
            self.weights_ = (weights / (weights.sum() + 10 * EPS) + EPS)
        if 'c' in params:
            covar_mstep_func = _covar_mstep_funcs[self.covariance_type]
            self.covars_ = covar_mstep_func(
                self, X, responsibilities, weighted_X_sum, inverse_weights,
                min_covar)
        return weights

    def _n_parameters(self):
        """Return the number of free parameters in the model."""
        ndim = self.means_.shape[1]
        if self.covariance_type == 'full':
            cov_params = self.n_components * ndim * (ndim + 1) / 2.
        elif self.covariance_type == 'diag':
            cov_params = self.n_components * ndim
        elif self.covariance_type == 'tied':
            cov_params = ndim * (ndim + 1) / 2.
        elif self.covariance_type == 'spherical':
            cov_params = self.n_components
        mean_params = ndim * self.n_components
        return int(cov_params + mean_params + self.n_components - 1)

    def bic(self, X, w=None):
        """Bayesian information criterion for the current model fit
        and the proposed data

        Parameters
        ----------
        X : array of shape(n_samples, n_dimensions)

        Returns
        -------
        bic: float (the lower the better)
        """
        return (-2 * self.score(X, w).sum() +
                self._n_parameters() * np.log(X.shape[0]))

    def aic(self, X, w=None):
        """Akaike information criterion for the current model fit
        and the proposed data

        Parameters
        ----------
        X : array of shape(n_samples, n_dimensions)

        Returns
        -------
        aic: float (the lower the better)
        """
        return - 2 * self.score(X, w).sum() + 2 * self._n_parameters()


#########################################################################
# some helper routines
#########################################################################


def _log_multivariate_normal_density_diag(X, means, covars):
    """Compute Gaussian log-density at X for a diagonal model"""
    n_samples, n_dim = X.shape
    lpr = -0.5 * (n_dim * np.log(2 * np.pi) + np.sum(np.log(covars), 1)
                  + np.sum((means ** 2) / covars, 1)
                  - 2 * np.dot(X, (means / covars).T)
                  + np.dot(X ** 2, (1.0 / covars).T))
    return lpr


def _log_multivariate_normal_density_spherical(X, means, covars):
    """Compute Gaussian log-density at X for a spherical model"""
    cv = covars.copy()
    if covars.ndim == 1:
        cv = cv[:, np.newaxis]
    if covars.shape[1] == 1:
        cv = np.tile(cv, (1, X.shape[-1]))
    return _log_multivariate_normal_density_diag(X, means, cv)


def _log_multivariate_normal_density_tied(X, means, covars):
    """Compute Gaussian log-density at X for a tied model"""
    cv = np.tile(covars, (means.shape[0], 1, 1))
    return _log_multivariate_normal_density_full(X, means, cv)


def _log_multivariate_normal_density_full(X, means, covars, min_covar=1.e-7):
    """Log probability for full covariance matrices."""
    n_samples, n_dim = X.shape
    nmix = len(means)
    log_prob = np.empty((n_samples, nmix))
    for c, (mu, cv) in enumerate(zip(means, covars)):
        try:
            cv_chol = linalg.cholesky(cv, lower=True)
        except linalg.LinAlgError:
            # The model is most probably stuck in a component with too
            # few observations, we need to reinitialize this components
            try:
                cv_chol = linalg.cholesky(cv + min_covar * np.eye(n_dim),
                                          lower=True)
            except linalg.LinAlgError:
                raise ValueError("'covars' must be symmetric, "
                                 "positive-definite")

        cv_log_det = 2 * np.sum(np.log(np.diagonal(cv_chol)))
        cv_sol = linalg.solve_triangular(cv_chol, (X - mu).T, lower=True).T
        log_prob[:, c] = - .5 * (np.sum(cv_sol ** 2, axis=1) +
                                 n_dim * np.log(2 * np.pi) + cv_log_det)

    return log_prob


def _validate_covars(covars, covariance_type, n_components):
    """Do basic checks on matrix covariance sizes and values
    """
    from scipy import linalg
    if covariance_type == 'spherical':
        if len(covars) != n_components:
            raise ValueError("'spherical' covars have length n_components")
        elif np.any(covars <= 0):
            raise ValueError("'spherical' covars must be non-negative")
    elif covariance_type == 'tied':
        if covars.shape[0] != covars.shape[1]:
            raise ValueError("'tied' covars must have shape (n_dim, n_dim)")
        elif (not np.allclose(covars, covars.T)
              or np.any(linalg.eigvalsh(covars) <= 0)):
            raise ValueError("'tied' covars must be symmetric, "
                             "positive-definite")
    elif covariance_type == 'diag':
        if len(covars.shape) != 2:
            raise ValueError("'diag' covars must have shape "
                             "(n_components, n_dim)")
        elif np.any(covars <= 0):
            raise ValueError("'diag' covars must be non-negative")
    elif covariance_type == 'full':
        if len(covars.shape) != 3:
            raise ValueError("'full' covars must have shape "
                             "(n_components, n_dim, n_dim)")
        elif covars.shape[1] != covars.shape[2]:
            raise ValueError("'full' covars must have shape "
                             "(n_components, n_dim, n_dim)")
        for n, cv in enumerate(covars):
            if (not np.allclose(cv, cv.T)
                    or np.any(linalg.eigvalsh(cv) <= 0)):
                raise ValueError("component %d of 'full' covars must be "
                                 "symmetric, positive-definite" % n)
    else:
        raise ValueError("covariance_type must be one of " +
                         "'spherical', 'tied', 'diag', 'full'")


def distribute_covar_matrix_to_match_covariance_type(
        tied_cv, covariance_type, n_components):
    """Create all the covariance matrices from a given template"""
    if covariance_type == 'spherical':
        cv = np.tile(tied_cv.mean() * np.ones(tied_cv.shape[1]),
                     (n_components, 1))
    elif covariance_type == 'tied':
        cv = tied_cv
    elif covariance_type == 'diag':
        cv = np.tile(np.diag(tied_cv), (n_components, 1))
    elif covariance_type == 'full':
        cv = np.tile(tied_cv, (n_components, 1, 1))
    else:
        raise ValueError("covariance_type must be one of " +
                         "'spherical', 'tied', 'diag', 'full'")
    return cv


def _covar_mstep_diag(gmm, X, responsibilities, weighted_X_sum, norm,
                      min_covar):
    """Performing the covariance M step for diagonal cases"""
    avg_X2 = np.dot(responsibilities.T, X * X) * norm
    avg_means2 = gmm.means_ ** 2
    avg_X_means = gmm.means_ * weighted_X_sum * norm
    return avg_X2 - 2 * avg_X_means + avg_means2 + min_covar


def _covar_mstep_spherical(*args):
    """Performing the covariance M step for spherical cases"""
    cv = _covar_mstep_diag(*args)
    return np.tile(cv.mean(axis=1)[:, np.newaxis], (1, cv.shape[1]))


def _covar_mstep_full(gmm, X, responsibilities, weighted_X_sum, norm,
                      min_covar):
    """Performing the covariance M step for full cases"""
    # Eq. 12 from K. Murphy, "Fitting a Conditional Linear Gaussian
    # Distribution"
    n_features = X.shape[1]
    cv = np.empty((gmm.n_components, n_features, n_features))
    for c in range(gmm.n_components):
        post = responsibilities[:, c]
        mu = gmm.means_[c]
        diff = X - mu
        with np.errstate(under='ignore'):
            # Underflow Errors in doing post * X.T are  not important
            avg_cv = np.dot(post * diff.T, diff) / (post.sum() + 10 * EPS)
        cv[c] = avg_cv + min_covar * np.eye(n_features)
    return cv


def _covar_mstep_tied(gmm, X, responsibilities, weighted_X_sum, norm,
                      min_covar):
    # Eq. 15 from K. Murphy, "Fitting a Conditional Linear Gaussian
    # Distribution"
    avg_X2 = np.dot(X.T, X)
    avg_means2 = np.dot(gmm.means_.T, weighted_X_sum)
    out = avg_X2 - avg_means2
    out *= 1. / X.shape[0]
    out.flat[::len(out) + 1] += min_covar
    return out

_covar_mstep_funcs = {'spherical': _covar_mstep_spherical,
                      'diag': _covar_mstep_diag,
                      'tied': _covar_mstep_tied,
                      'full': _covar_mstep_full,
                      }


def load_model(data, random_state=None):
    import operator
    import numpy

    total_weight = data.attrs["total_weight"]

    # Count the number of components necessary
    K_total = 0
    for name, group in data.items():
        if not name.startswith("class") or group.attrs["weight"] == 0:
            continue
        K_total += group.attrs["K_best"]

    ndim = data.attrs["ndim"]

    weights = numpy.empty(K_total)
    means = numpy.empty((K_total, ndim))
    covars = numpy.empty((K_total, ndim, ndim))

    i = 0
    for name, group in data.items():
        if not name.startswith("class") or group.attrs["weight"] == 0:
            continue

        K = int(group.attrs["K_best"])
        i_next = i + K

        subgroup = group[str(K)]

        group_weight = group.attrs["weight"] / total_weight

        weights[i:i_next] = group_weight*subgroup["weights"]
        means[i:i_next] = subgroup["means"]
        covars[i:i_next] = subgroup["covars"]

        i = i_next

    model = WeightedGMM(
        n_components=len(weights),
        covariance_type="full",
        random_state=random_state,
    )
    model.weights_ = weights
    model.means_ = means
    model.covars_ = covars

    return model



#
# efficient_save_sklearn_gp.py
#
# GOAL
#   - sklearn gp objects I use always have a simple form
#   - this saves a json file holding *strings* describing the kernel, AND the X_train_ and y_train_ data used in the GP
#   - we *may* still  use a pickle for now, but it is *much* more efficient -- we're just saving the data we need
#   - this lets me change the data format to make it *portable*

import json

import numpy as np

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, WhiteKernel, ConstantKernel as C

def gp2json(gp):
    """
    Converts gp object to (a) json object and (b) data for X_train_ and y_train
    """
    out_dict = {}
    out_dict['kernel'] = str(gp.kernel_)
    out_dict['kernel_params'] = {}
    dict_params = gp.kernel_.get_params()
    for name in dict_params:
        out_dict['kernel_params'][name] = str(dict_params[name])   # gives me ability to set more parameters in greater detail
    out_dict['y_train_std'] = str(gp._y_train_std)
    out_dict['y_train_mean'] = str(gp._y_train_mean)
    return [out_dict, gp.X_train_, gp.y_train_,gp.alpha_]

def export_gp_compact(fname_base,gp):
    fname_json = fname_base+".json"
    fname_dat_X = fname_base+"_X.dat"
    fname_dat_y = fname_base+"_y.dat"
    fname_dat_alpha = fname_base+"_alpha.dat"
    my_json, my_X,my_y, my_alpha = gp2json(gp) 
    np.savetxt(fname_dat_X,my_X)
    np.savetxt(fname_dat_y,my_y)
    np.savetxt(fname_dat_alpha,my_alpha)
    with open(fname_json,'w') as f:
        json.dump(my_json, f)
    return None


def load_gp(fname_base):
    kernel=None
    with open(fname_base+".json",'r') as f:
        my_json = json.load(f)
    my_X = np.loadtxt(fname_base+"_X.dat")
    my_y = np.loadtxt(fname_base+"_y.dat")
    my_alpha = np.loadtxt(fname_base+"_alpha.dat")
    dict_params = my_json['kernel_params']
    eval("kernel = "+my_json['kernel'])
    gp = GaussianProcessRegressor(kernel=kernel,n_restarts_optimizer=0)
    gp.kernel_ = kernel
    dict_params_eval = {}
    for name in dict_params:
        if not('length' in name   or 'constant' in name):
            continue
        if name =="k2__k2__length_scale":
            one_space = ' '.join(dict_params[name].split())
            dict_params_eval[name] = eval(one_space.replace(' ',','))
        else:
            dict_params_eval[name] = eval(dict_params[name])
    gp.kernel_.set_params(dict_params_eval)
    gp.X_train_ = my_X
    gp.y_train_ = my_y
    gp.alpha_ = my_alpha
    gp._y_train_std = float(my_json['y_train_std'])
    gp._y_train_mean = float(my_json['y_train_mean'])
    return gp

                            



#! /usr/bin/env python
#
# GOAL
#    - evaluate overlap between two waveforms (two inj xml) in H, L
#    - prints out mismatch in H, L.
#    - no time maximization by default.  (We can do it)
#
# COMPARE TO
#   - util_NRCompareSimulations.py
#   - util_ManualOverlapGrid
#   - util_LALCompareApproximantsOnSamples.py
#
# DEMOS
#     * Compare two approximants with each other, on the same parameters
#     python util_CompareWaveformsInDetectors.py --inj1 inj.xml.gz --inj2 inj.xml.gz --approx SEOBNRv2 --approx2 SEOBNRv4
#     
#    * Compare one single waveform (e.g., NR best fits) with many others (e.g., output of precessing parameters)
#      python util_CompareWaveformsInDetectors.py --inj inj1.xml.gz --inj2 maxpt.xml.gz --approx SEOBNRv2 --group2 Sequence-SXS-All --param2 1 --use-provided-strain2
#
#   * Compare two NR waveforms that have the same parameters

from __future__ import print_function

import argparse
import numpy as np

import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import sys

import NRWaveformCatalogManager3 as nrwf
from matplotlib import pyplot as plt


parser = argparse.ArgumentParser()
parser.add_argument("--maximize",action='store_true',help="Perform overlap instead of faithfulness test. Important if inconsistent times")
parser.add_argument("--inj",default=None,type=str,help="Required. Arguments for 1. If used, NR will override intrinsic parameters")
parser.add_argument("--inj2",default=None,type=str,help="Required. Arguments for 2. If used, NR will override intrinsic parameters")
parser.add_argument("--group", default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--param", default="1",help="Parameter value")
parser.add_argument("--group2", default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--param2", default="1",help="Parameter value")
parser.add_argument("--align-22",action='store_true',help='Find the optimal time and phase for alignment for the 22 mode [then halt]')
parser.add_argument("--use-interpolated-overlap",action='store_true', help="The overlap uses a local quadratic approximation near the peak, not just the maximum value")
parser.add_argument("--use-perturbative-extraction", action='store_true')
parser.add_argument("--use-perturbative-extraction-full", action='store_true')
parser.add_argument("--use-provided-strain", action='store_true')
parser.add_argument("--use-provided-strain2", action='store_true')
parser.add_argument("--use-spec-lev",type=int,default=5)
parser.add_argument("--use-hybrid",action='store_true')
parser.add_argument("--use-hybrid-method",default='taper_add')
parser.add_argument("--use-hybrid2",action='store_true')
parser.add_argument("--use-hybrid2-method",default='taper_add')
parser.add_argument("--approx", default="EOBNRv2HM", help="approximant to use for comparison")
parser.add_argument("--approx2", default="EOBNRv2HM", help="approximant to use for comparison")
parser.add_argument("--lmax",default=2,type=int)
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen",type=int,default=16,help="Window time")
parser.add_argument("--fref",default=None,type=float,help="Reference frqeuency (assumed equal to fmin)")
parser.add_argument("--fmin",default=20,type=float,help="Minimum frequency for overlap integral -- NOT necessarily tied to tempalte starting frequencies!")
parser.add_argument("--fmin-template1",default=None,type=float,help="Override template frequency 1 (e.g., for NR XML files)")
parser.add_argument("--fmin-template2",default=None,type=float,help="Override template frequency 1 (e.g., for NR XML files)")
parser.add_argument("--fmax",default=2000,type=float,help="Maximum frequency in Hz, used for PSD integral.")
parser.add_argument("--psd-file",default=None,action='append',help="PSD file")
parser.add_argument("--psd",type=str,default="SimNoisePSDaLIGOZeroDetHighPower",help="psd name (attribute in lalsimulation).  SimNoisePSDiLIGOSRD, lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsimutils.Wrapper_AdvLIGOPsd, .SimNoisePSDiLIGOSRD... ")
parser.add_argument("--fname-output", default="comparison_output.dat",type=str)
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--save-plots",action="store_true")
parser.add_argument("--tref",default=None,type=float)

opts = parser.parse_args()


###
### Spacing settings, for overlaps
###


lmax = opts.lmax
T_window = opts.seglen
df = 1./T_window
fmin =opts.fmin  # default for now
# if opts.fmin_template1:
#     fmin = np.min([fmin, opts.fmin_template1])
# if opts.fmin_template2:
#     fmin = np.min([fmin, opts.fmin_template2])
fmaxSNR=1700
analyticPSD_Q=True

psd_dict = {}
ifo_list = []

psd=lalsim.SimNoisePSDaLIGOZeroDetHighPower
if opts.psd_file:
    analyticPSD_Q=False

    for inst, psdf in map(lambda c: c.split("="), opts.psd_file):
        if opts.verbose: 
            print("Reading PSD for instrument %s from %s" % (inst, psdf))
        psd_dict[inst] = lalsimutils.load_resample_and_clean_psd(psdf, inst, df)

    ifo_list = psd_dict.keys()

elif opts.psd and hasattr(lalsim, opts.psd):
    ifo_list = ['H1','L1']
    psd = getattr(lalsim, opts.psd)


fNyq = opts.srate/2.
IP=None
IP_list = {}
if opts.maximize:
    for ifo in ifo_list:
        IP_list[ifo]= lalsimutils.ComplexOverlap(fNyq=fNyq,deltaF=df,analyticPSD_Q=analyticPSD_Q,psd=psd_dict[ifo],fMax=opts.fmax,fLow=fmin)
else:
    for ifo in ifo_list:
        IP_list[ifo] = lalsimutils.ComplexIP(fNyq=fNyq,deltaF=df,analyticPSD_Q=analyticPSD_Q,psd=psd_dict[ifo],fMax=opts.fmax,fLow=fmin)

###
### Load injection XML
###

if opts.inj is None:
    print(" --inj required")
    sys.exit(0)
print(" Reading injection file 1 for comparison ", opts.inj)
P1_list = lalsimutils.xml_to_ChooseWaveformParams_array(opts.inj)
nlines1  = len(P1_list)
print(" Read  ", nlines1, " injections")

if nlines1 < 1:
    print(" No data in ", opts.inj)

tref = float( (P1_list[0]).tref ) # default
if not(opts.tref is None):
    tref = opts.tref

if opts.inj2 is None:
    print(" --inj2 required")
    sys.exit(0)
print(" Reading injection file 1 for comparison ", opts.inj2)
P2_list = lalsimutils.xml_to_ChooseWaveformParams_array(opts.inj2)
nlines2  = len(P2_list)

if nlines2 < 1:
    print(" No data in ", opts.inj2)




###
### Define functions to generate waveforms
###

def get_hF1(indx,ifo):
    global opts
    P = P1_list[indx % nlines1].manual_copy() # looping
    if not (opts.fmin_template1 is None):
        P.fmin = opts.fmin_template1
    if not (opts.fref is None):
        P.fref = opts.fref
    else:
        P.ref = P.fmin
    P.radec =True
    P.tref = P1_list[indx%nlines1].tref # copy, this is an allocated object
    P.deltaF = df
    P.deltaT = 1./opts.srate
    P.detector = ifo
    P.approx = lalsim.GetApproximantFromString(opts.approx)  # override the XML. Can screw you up (ref spins)
    # if P.approx == lalsim.IMRPhenomPv2:
    #     phiJL_now = P.extract_param('phiJL')
    #     P.assign_param('phiJL', phiJL_now-np.pi/2)  # Estimate
#        P.fref = 100
    if opts.verbose:
        P.print_params()
    hF = lalsimutils.non_herm_hoff(P)
    return hF

def get_hF2(indx,ifo):
    global opts
    P = P2_list[indx % nlines2].manual_copy() # looping
    if not (opts.fmin_template2 is None):
        P.fmin = opts.fmin_template2
    if not (opts.fref is None):
        P.fref = opts.fref
    else:
        P.ref = P.fmin
    P.radec =True
    P.tref = P2_list[indx%nlines2].tref # copy, this is an allocated object
    P.deltaF = df
    P.deltaT = 1./opts.srate
    P.detector = ifo
    P.approx = lalsim.GetApproximantFromString(opts.approx2)  # override the XML. Can screw you up
    if P.approx == lalsim.IMRPhenomPv2:
        cosbeta =np.cos(P.extract_param('beta'))
        my_phase = np.pi -np.pi/2
        dt = 20*(P.m1+P.m2)/lal.MSUN_SI * lalsimutils.MsunInSec  # ad hoc factor .. there is apparently a timeshift of Pv2 relative to PD of about 20 M
        P.tref += dt
        phiJL_now = P.extract_param('phiJL')
        psi_now = P.psi
        P.assign_param('phiJL', phiJL_now-my_phase)  # Estimate
        P.psi  = psi_now   # fix L alignment, not J
        P.phiref += my_phase 
        P.phiref += np.pi/2 
    if opts.verbose:
        P.print_params()
    hF = lalsimutils.non_herm_hoff(P)
    return hF


return_hF1 = get_hF1
return_hF2 = get_hF2

###
### Load NR (as needed) and redefine generation functions
###


group = opts.group
param = None
wfP =  None
if group in nrwf.internal_ParametersAreExpressions.keys():
    if not nrwf.internal_ParametersAreExpressions[group]:
        param = opts.param
    else:
        param = eval(str(opts.param))
    if opts.use_spec_lev:
            if group == "Sequence-SXS-All":
                nrwf.internal_FilenamesForParameters[group][param] =nrwf.internal_FilenamesForParameters[group][param].replace("Lev5", "Lev" +str(opts.use_spec_lev))
                print(" OVERRIDE OF SXS LEVEL : ", nrwf.internal_FilenamesForParameters[group][param])
    wfP = nrwf.WaveformModeCatalog(opts.group, param, clean_initial_transient=True,clean_final_decay=True, shift_by_extraction_radius=True, lmax=lmax,align_at_peak_l2_m2_emission=True,perturbative_extraction=opts.use_perturbative_extraction,perturbative_extraction_full=opts.use_perturbative_extraction_full,use_provided_strain=opts.use_provided_strain,reference_phase_at_peak=True,quiet=True)
    wfP.P.fmin  = opts.fmin

    print(" WARNING: NR orientation parameters differ, so the waveform CANNOT be identical and will generally differ in at least orbital phase ")

    def get_hF1_NR(indx,ifo):
        P_here = P1_list[indx % nlines1]

        wfP.P.radec=True
        wfP.P.m1 = P_here.m1
        wfP.P.m2 = P_here.m2

        wfP.P.dist = P_here.dist
        wfP.P.incl = P_here.incl
        wfP.P.theta = P_here.theta
        wfP.P.phi = P_here.phi
        wfP.P.tref = P_here.tref
        wfP.P.psi = P_here.psi
        wfP.P.phiref = P_here.phiref


        wfP.P.fmin = opts.fmin
        wfP.P.deltaF = df
        wfP.P.deltaT = 1./opts.srate
        wfP.P.detector = ifo
        wfP.P.approx = lalsim.GetApproximantFromString(opts.approx)  # override the XML. Can screw you up (ref spins)

        if opts.verbose:
            wfP.P.print_params()

        hF = wfP.non_herm_hoff()
        return hF
    
    return_hF1 = get_hF1_NR


group2 = opts.group2
if group2 in nrwf.internal_ParametersAreExpressions.keys():
    if not nrwf.internal_ParametersAreExpressions[group2]:
        param2 = opts.param2
    else:
        param2 = eval(str(opts.param2))
    # Check for SXS 
    if group2 == "Sequence-SXS-All":
        nrwf.internal_FilenamesForParameters[group2][param2] =nrwf.internal_FilenamesForParameters[group2][param2].replace("Lev5", "Lev" +str(opts.use_spec_lev))
        print(" OVERRIDE OF SXS LEVEL : ", nrwf.internal_FilenamesForParameters[group2][param2])
    if opts.verbose:
        print("Importing ", group, param , " and ", group2, param2)
    wfP2 = nrwf.WaveformModeCatalog(opts.group2, param2, clean_initial_transient=True,clean_final_decay=True, shift_by_extraction_radius=True, lmax=lmax,align_at_peak_l2_m2_emission=True,perturbative_extraction=opts.use_perturbative_extraction,use_provided_strain=opts.use_provided_strain2,reference_phase_at_peak=True,quiet=True)
    wfP2.P.fmin  = opts.fmin

    print(" WARNING: NR orientation parameters differ, so the waveform CANNOT be identical and will generally differ in at least orbital phase ")
    def get_hF2_NR(indx,ifo):
        P_here = P2_list[indx % nlines2]

        wfP2.P.radec=True
        wfP2.P.m1 = P_here.m1
        wfP2.P.m2 = P_here.m2

        wfP2.P.dist = P_here.dist
        wfP2.P.incl = P_here.incl
        wfP2.P.theta = P_here.theta
        wfP2.P.phi = P_here.phi
        wfP2.P.tref = P_here.tref
        wfP2.P.psi = P_here.psi
        wfP2.P.phiref = P_here.phiref


        wfP2.P.fmin = opts.fmin
        wfP2.P.deltaF = df
        wfP2.P.deltaT = 1./opts.srate
        wfP2.P.detector = ifo
        wfP2.P.approx = lalsim.GetApproximantFromString(opts.approx2)  # override the XML. Can screw you up (ref spins)
        if opts.verbose:
            wfP2.P.print_params()

        hF = wfP2.non_herm_hoff()

        return hF

    
    return_hF2 = get_hF2_NR








###
### Loop and carry out overlap
###

n_evals = np.max([nlines1,nlines2])

dat_out =[]

for indx in np.arange(n_evals):
  if opts.verbose:
      print(indx)
  if True:
#  try:
    line = []
    print(P1_list[indx].extract_param('thetaJN'), P1_list[indx].phiref, P1_list[indx].extract_param('beta'), end=' ')
    for ifo in ifo_list: #,'V1']:
        IP = IP_list[ifo]
        hF1 = return_hF1(indx,ifo)
        nm1 = IP.norm(hF1)
        hF2 = return_hF2(indx,ifo)
        nm2 = IP.norm(hF2)
        val = np.abs(IP.ip(hF1,hF2,include_epoch_differences=True)/nm1/nm2)  # correct for timeshift issues, if any
        print(" Epoch test ", hF1.epoch, hF2.epoch)
        line.append(val)
#        print ifo, IP.ip(hF1,hF2)/nm1/nm2, IP.ip(hF1,hF2),nm1, nm2
        print(val, end=' ')
        if opts.save_plots:
            if opts.verbose:
                print(" --- Saving plot for ", ifo, " ----")
            label1 = opts.approx
            label2 = opts.approx2
            hT1 = lalsimutils.DataInverseFourier(hF1)
            hT2 = lalsimutils.DataInverseFourier(hF2)
            if not (opts.group is None):
                npts = hT1.data.length
                T_wave =npts*hT1.deltaT
                hT1 = lalsimutils.DataRollTime(hT1,-0.5*T_wave/2)
                if opts.verbose:
                    print(" Epoch1 ", float(hT1.epoch))
                label1 = group+":"+param
            if not (opts.group2 is None):
                if opts.verbose:
                    print(" ---> Rolling to fix FT centering <-- ")
                npts = hT2.data.length
                T_wave =npts*hT2.deltaT
                hT2 = lalsimutils.DataRollTime(hT2,-0.5*T_wave/2)
                if opts.verbose:
                    print(" Epoch2 ", hT2.epoch)
                label2 = group2+":"+param2
            tvals = lalsimutils.evaluate_tvals(hT1) 
            tvals2 = lalsimutils.evaluate_tvals(hT2) 
#            tref = float(hT1.epoch)
            plt.plot(tvals -tref,np.real(hT1.data.data),'r',label=label1)
            plt.plot(tvals2 -tref,np.real(hT2.data.data),'g',label=label2)
            plt.legend()
            plt.savefig(opts.fname_output+"_fig_"+str(indx)+"_"+ifo+".png"); 
            
            # Estimate peak location, so plot can be focused on detail
            tmax_offset =(tvals- tref)[ np.argmax( np.abs(hT1.data.data))]  
                
            plt.xlim(tmax_offset-0.15, tmax_offset+0.05)  # NR scale, focus on merger
            plt.savefig(opts.fname_output+"_fig_"+str(indx)+"_"+ifo+"_detail.png"); 
            plt.clf()
    print()
    dat_out.append(line)

#  except:
  else:
      print(" Skipping ", indx)

np.savetxt(opts.fname_output, np.array(dat_out))



#! /usr/bin/env python
"""
Author: Ethan Payne (ethan.payne@ligo.org)

Calibration marginalization of any generic set of samples. Note that this needs to be used within reason.
The calibration marginalization generates N_ifo matrices of size N_freq_bins by N_calibration_curves.
This can be very memory intensive.

We assume that the likelihood required is only the GWTransientLikelihood with and without calibration marginalization.
No other marginalizations are included. The calibration marginalized likelihood can work with both distance and phase
marginalization if desired.

Also note that the rejection sampling used can be quite lossy - even though it does not significantly change the
posterior distribution.

Inputs
------
posterior_sample_file: Bilby result .json or .dat file
    Either a Bilby result file or a .dat file with headers. Contains the posterior samples of the distribution to be
    used in the reweighting calculation.
data_dump_file: .pickle file
    File produced from bilby_pipe_generation config.ini. This gives an easy way to import in all the required data, and
    in the case of bilby_pipe runs, ensure the same data is used between the original PE and the reweighting.
number_of_calibration_curves: int
    Number of draws of the calibration model which we marginalize over.
    Default: 100.
reevaluate_likelihood: bool
    Sets whether we want to also reevaluate the likelihood without calibration uncertainty. This can ensure that the
    likelihood definitions are consistent (i.e. log_likelihood or log_likelihood - log_noise_evidence).
    Default: True

Output
------
reweighted_posterior_samples.dat:
    Data file with the rejection sampled posterior samples from the supplied posterior file
weights.dat:
    File with the weight for each sample in the supplied posterior file. This allows the user to plot results using the
    the weights as opposed to the rejection sampling posterior samples.
"""

import os
import sys

import argparse
from copy import deepcopy, copy

import numpy as np
import pandas as pd
import pickle

import bilby
import bilby_pipe

# So I can import the RIFT source while not making this some setup.py-able package
# TODO this should not be a hardcoded path!

import RIFT.calmarg.rift_source as rift_source

parser = argparse.ArgumentParser(description='calibration marginalization via reweighting of posterior samples')
parser.add(
    "--posterior_sample_file", default=None,
    help="Bilby result file or text file with posterior samples to reweight")
parser.add(
    "--data_dump_file", default=None,
    help="Bilby pipe data_dump file. This stores all the relevant data for PE")
parser.add(
    "--number_of_calibration_curves", type=int, default=1000,
    help="Number of calibration curve realizations to use to reweight samples.")
parser.add(
    "--time_marginalization", type=bool, default=False,
    help="Uses time maginalization when evaluating the likelihoods. Defaults to False")
parser.add(
    "--phase_marginalization", type=bool, default=False,
    help="Uses phase maginalization when evaluating the likelihoods. Defaults to False")
parser.add(
    "--reevaluate_likelihood", type=bool, default=True,
    help="Reevaluates the likelihood of the given samples using the likelihood from the data_dump.")
parser.add(
    "--use_nested_samples", type=bool, default=False,
    help="Use nested samples with their weights to generate the samples. This significantly improves efficiency.")
parser.add(
    "--time_marginalization_interval", default=0.1, type=float,
    help="Interval over which the time marginalization is undertaken")
parser.add(
    "--use_rift_samples", type=bool, default=False,
    help="Uses a different waveform function if using RIFT samples. This matches the phase definitions in RIFT")
parser.add(
    "--h_method", type=str, default='hlmoft',
    help="For RIFT, uses two different options for h(t) reconstruction ")
parser.add("--internal-waveform-fd-L-frame",action='store_true',help='If true, passes extra_waveform_kwargs = {fd_L_frame=True} to lalsimutils hlmoft. Impacts outputs of ChooseFDWaveform calls only.')
parser.add("--internal-waveform-fd-no-condition",action='store_true',help='If true, adds extra_waveform_kwargs = {no_condition=True} to lalsimutils hlmoft. Impacts outputs of ChooseFDWaveform calls only. Provided to enable controlled tests of conditioning impact on PE')
parser.add("--fmin", default=None, type=float)
parser.add("--l-max", default=4, type=int)
parser.add("--start_index", default=None, type=int)
parser.add("--end_index", default=None, type=int)

args = parser.parse_args()

with open(args.data_dump_file, "rb") as data_file:
    data = pickle.load(data_file)

start_index = args.start_index
end_index = args.end_index

# read in the posterior samples for reweighting
if args.posterior_sample_file.split(".")[-1] == 'json':
    result = bilby.core.result.read_in_result(args.posterior_sample_file)
    if end_index is not None:
        if end_index > len(result.posterior):
            end_index = len(result.posterior)
    result.posterior = result.posterior.iloc[start_index:end_index].reset_index(drop=True)
elif (args.posterior_sample_file.split(".")[-1] == 'txt') or (args.posterior_sample_file.split(".")[-1] == 'dat'):
    result = bilby.core.result.Result()
    result.posterior = pd.DataFrame(np.genfromtxt(args.posterior_sample_file, names=True))
    if end_index is not None:
        if end_index > len(result.posterior):
            end_index = len(result.posterior)
    result.posterior = result.posterior.iloc[start_index:end_index].reset_index(drop=True)
    result.meta_data = {}

    if args.use_rift_samples:
        result.posterior  = result.posterior.drop(columns=['lnL','ps'])
        result.posterior['p'] = np.log(result.posterior['p'])  # not sure if used, but if so define correctly
        key_swap_dict = {'m1':'mass_1', 'm2':'mass_2', 'a1x':'spin_1x', 'a1y':'spin_1y', 'a1z':'spin_1z',
            'a2x':'spin_2x', 'a2y':'spin_2y', 'a2z':'spin_2z', 'incl':'iota', 'time':'geocent_time',
        'phiorb':'phase', 'p':'log_prior', 'distance':'luminosity_distance', 'lambda1':'lambda_1', 'lambda2':'lambda_2'}

        names = list(result.posterior.keys())  # dangerous to have iterator tied to changing structure
        for old_key in names:
            if old_key in key_swap_dict:
                result.posterior[key_swap_dict[old_key]] = result.posterior[old_key]
                del result.posterior[old_key]
        # add tides if not present
        if not('lambda_1' in result.posterior):
            print(" Populating empty tidal params to avoid hang")
            if end_index:
              result.posterior['lambda_1'] = np.zeros(end_index-start_index)
              result.posterior['lambda_2'] = np.zeros(end_index-start_index)
            else:
              result.posterior['lambda_1'] = np.zeros(len(result.posterior['mass_1']))
              result.posterior['lambda_2'] = np.zeros(len(result.posterior['mass_1']))

outdir = os.path.dirname(os.path.abspath(args.posterior_sample_file))

ifos = data.interferometers
time_marginalization_interval = args.time_marginalization_interval

spline_calibration_envelope_dict = bilby_pipe.utils.convert_string_to_dict(
                data.meta_data['command_line_args']['spline_calibration_envelope_dict'])
ifos_for_reweighting = deepcopy(ifos)
for ifo in ifos: # removes any model for the calibration that was set up in the file
    ifo.calibration_model = bilby.gw.calibration.Recalibrate()

if args.time_marginalization:
    result.posterior['geocent_time'] = ifos.start_time * np.ones(len(result.posterior))
    if 'time_jitter' not in result.posterior.keys():
        result.posterior['time_jitter'] = np.zeros(len(result.posterior))

if args.phase_marginalization:
    if not args.use_rift_samples:
        result.posterior['phase'] = np.zeros(len(result.posterior))

# Setting up the waveform generator using the data dump features
waveform_arguments = dict(
    reference_frequency=data.meta_data['command_line_args']['reference_frequency'],
    minimum_frequency=args.fmin,
    waveform_approximant=data.meta_data['command_line_args']['waveform_approximant'],
    sampling_frequency=ifos.sampling_frequency,
    h_method=args.h_method)

extra_waveform_kwargs={}
if args.internal_waveform_fd_L_frame:
    extra_waveform_kwargs = {'fd_L_frame':True}
if args.internal_waveform_fd_no_condition:
    extra_waveform_kwargs['no_condition'] = True
waveform_arguments['extra_waveform_kwargs'] = extra_waveform_kwargs

if args.use_rift_samples:
    waveform_arguments['Lmax'] = args.l_max
#    waveform_arguments['waveform_approximant'] = 'SEOBNRv4PHM'
    wf_func = rift_source.RIFT_lal_binary_black_hole
else:
    wf_func = eval('bilby.gw.source.'+data.meta_data['command_line_args']['frequency_domain_source_model'])

waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            frequency_domain_source_model=wf_func,
            sampling_frequency=ifos.sampling_frequency,
            duration=ifos.duration,
            start_time=ifos.start_time,
            waveform_arguments=waveform_arguments)

# Setting up the calibration priors given the spline files for each interferometer
# This is used to generate draws of the calibration uncertainty from their prior
priors = bilby.core.prior.PriorDict()
for ifo in ifos_for_reweighting:
    calibration_file_path = f'{spline_calibration_envelope_dict[ifo.name]}'
    ifo_calibration_priors = bilby.gw.prior.CalibrationPriorDict.from_envelope_file(
        calibration_file_path, ifo.minimum_frequency, ifo.maximum_frequency, 10, ifo.name)

    # TODO FOR DEBUGGING PURPOSES
    # for key in ifo_calibration_priors.keys():
    #     if 'amp' in key:
    #         ifo_calibration_priors[key] = bilby.core.prior.DeltaFunction(peak=0)
    #     elif 'phase' in key:
    #         ifo_calibration_priors[key] = bilby.core.prior.DeltaFunction(peak=0)

    priors.update(ifo_calibration_priors)


marg_priors = bilby.core.prior.PriorDict() # priors for the non-calibration likelihood

if args.time_marginalization:
    priors['geocent_time'] = bilby.core.prior.Uniform(
        float(data.meta_data['command_line_args']['trigger_time']) - time_marginalization_interval,
        float(data.meta_data['command_line_args']['trigger_time']) + time_marginalization_interval)

    marg_priors['geocent_time'] = bilby.core.prior.Uniform(
        float(data.meta_data['command_line_args']['trigger_time']) - time_marginalization_interval,
        float(data.meta_data['command_line_args']['trigger_time']) + time_marginalization_interval)

if args.phase_marginalization:
    priors['phase'] = bilby.core.prior.Uniform(0, 2 * np.pi)
    marg_priors['phase'] = bilby.core.prior.Uniform(0, 2 * np.pi)

# Setting up the output for the calibration draw files
calibration_lookup_table = {}
for ifo in ifos:
    calibration_lookup_table[ifo.name] =\
                        f'{outdir}/{ifo.name}_calibration_file.h5'

original_likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator,
    time_marginalization=args.time_marginalization,
    phase_marginalization=args.phase_marginalization, priors=marg_priors)

calibration_likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos_for_reweighting, waveform_generator=copy(waveform_generator),
    number_of_response_curves=args.number_of_calibration_curves, calibration_marginalization=True,
    priors=priors, time_marginalization=args.time_marginalization, phase_marginalization=args.phase_marginalization,
    calibration_lookup_table=calibration_lookup_table)

print(f'Log noise evidence: {original_likelihood.noise_log_likelihood()}')

# TODO have not yet implemented the parameter reconstruction in the case where
# time and calibration marginalization are used
if args.time_marginalization:
    conversion_function = None
else:
    conversion_function = bilby.gw.conversion.generate_all_bbh_parameters

# Setting up the resume and weight file names
resume_file = None
if (start_index != None) and (start_index != None):
    weights_file = f'{outdir}/weight_files/weights_s{start_index}e{end_index}.dat'
elif start_index != None:
    weights_file = f'{outdir}/weight_files/weights_s{start_index}eNone.dat'
elif end_index != None:
    weights_file = f'{outdir}/weight_files/weights_s0e{end_index}.dat'
else:
    resume_file = f'{outdir}/reweighting_resume.dat'
    weights_file = f'{outdir}/weights.dat'

if args.reevaluate_likelihood:
    result_reweighted, weights, _, _, _, _ = bilby.core.result.reweight(
        result, new_likelihood=calibration_likelihood, old_likelihood=original_likelihood,
        conversion_function=conversion_function, verbose_output=True,
        resume_file=resume_file, n_checkpoint=5000,
        use_nested_samples=args.use_nested_samples)
else:
    result_reweighted, weights, _, _, _, _ = bilby.core.result.reweight(
        result, new_likelihood=conversion_function,
        conversion_function=bilby.gw.conversion.generate_all_bbh_parameters, verbose_output=True,
        resume_file=resume_file, n_checkpoint=5000,
        use_nested_samples=args.use_nested_samples)

# Save the weights to a file, with the sample index in the name
if not os.path.exists(f'{outdir}/weight_files/'):
    os.makedirs(f'{outdir}/weight_files/')

np.savetxt(weights_file, weights)

if (start_index == None) and (end_index == None):
    result_reweighted.save_posterior_samples(filename=outdir+'/reweighted_posterior_samples.dat', outdir=outdir)



#! /usr/bin/env python
#
# GOAL
#   - take output of concatenated 'flatfile_points' files (=concatennated output of convert_output_format_ile2inference)
#   - assume concatenation is a FAIR DRAW FROM THE EXTRINSIC POSTERIOR 
#   - resample to produce extrinsic parameters and intrinsic parameters
#   
#
# ORIGINAL REFERENCES
#    - postprocess_1d_cumulative.py (based on 'flatfile-points.dat', original format for storing output)
#    - util_ConstructIntrinsicPosterior_GenericCoordinates.py
#
# EXAMPLES
#    convert_output_format_ile2inference zero_noise.xml.gz > flatfile-points.dat   
#    util_ResampleILEOutputWithExtrinsic.py --fname flatfile-points.dat  --n-output-samples 5


# Setup. 
import numpy as np
import lal
import RIFT.lalsimutils as lalsimutils
import bisect


import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--fname", default=None,help="Target file to read. Output of convert_output_format_ile2inference. Can merge, but must retain columns!")
parser.add_argument("--fname-out", default="output-with-extrinsic.xml.gz",help="Target output file, including extrinsic variables. No cosmology or source-frame masses...user must do that themselves.")
parser.add_argument("--save-P",default=0.,type=float,help="Not currently used")
parser.add_argument("--n-output-samples",default=2000,type=int)
opts = parser.parse_args()

samples = np.genfromtxt(opts.fname,names=True,invalid_raise=False)
lnLmax = np.max(samples["lnL"])
p = samples["p"]     # only prior!
ps = samples["ps"]  # sampling prior
Npts = samples["Npts"] # how many points in this iteration
weights = np.exp( samples["lnL"]-lnLmax) * (p/ps)/Npts   # each point will produce Npts underlying samples. Treat each *experiment* equally (instead of each pt)...suboptimal 

# if save_P is implemented, we will need to sort the weights, so we can downselect the unimportant ones

# Draw random numbers
p_threshold_size = opts.n_output_samples #np.min([5*opts.n_output_samples,len(weights)])
#p_thresholds =  np.random.uniform(low=0.0,high=1.0,size=p_threshold_size)#opts.n_output_samples)
#cum_sum  = np.cumsum(weights)
#cum_sum = cum_sum/cum_sum[-1]
#indx_list = map(lambda x : np.sum(cum_sum < x),  p_thresholds)  # this can lead to duplicates
indx_list = np.random.choice(np.arange(len(weights)), size=p_threshold_size, p=weights/np.sum(weights))
#print indx_list, samples["lnL"][indx_list], samples["distance"][indx_list], weights[indx_list], np.max(weights), samples["lnL"], lnLmax
#print  np.mean(samples["lnL"][indx_list]), np.mean(samples["distance"][indx_list]), weights[indx_list]


# downselect
samples_out = np.empty( len(indx_list), samples.dtype)
for param in samples.dtype.names:
    samples_out[param] = samples[param][indx_list]


# Write output (ascii)
# Vastly superior data i/o with pandas, but note no # used
import pandas
dframe = pandas.DataFrame(samples_out)
fname_out_ascii= opts.fname_out.replace(".xml.gz", "")+".dat"
dframe.to_csv(fname_out_ascii,sep=' ',header=' # '+' '.join(samples_out.dtype.names),index=False) # # is not pre-appended...just boolean

    



#! /usr/bin/env python
# util_CleanILE.py
#
#  Reads FILE (not stdin). Consolidates ILE entries for the same physical system.
#  Compare to: util_MassGriCoalesce.py


import sys
import os

import numpy as np
import RIFT.misc.weight_simulations as weight_simulations
import scipy.special

import fileinput
#import StringIO

data_at_intrinsic = {}

my_digits=7  # safety for high-SNR BNS


import argparse
parser = argparse.ArgumentParser(usage="util_CleanILE.py fname1.dat fname2.dat ... ")
parser.add_argument("fname",action='append',nargs='+')
parser.add_argument("--combination",default='average',help='average|sum, depends on if we treat them as independent trials of the same quantity or independent')
opts = parser.parse_args()

#print opts.fname
from pathlib import Path
for fname in opts.fname[0]: #sys.argv[1:]:
    fname  = Path(fname).resolve()
    if not( os.path.exists(fname)):  # skip symbolic links that don't resolve : important for .composite files
        continue
    if os.stat(fname).st_size==0:  # skip files of zero length
        continue
    sys.stderr.write(str(fname)+"\n")
#    data = np.loadtxt(fname)  # this will FAIL if we have a heterogeneous data source!  BE CAREFUL
    data = np.genfromtxt(fname,invalid_raise=False)  #  Protect against inhomogeneous data
    if len(data.shape)==1:
        data = np.array([data])
    for line in data:
      try:
        line = np.around(line, decimals=my_digits)
        if tuple(line[2:]) in data_at_intrinsic:
#            print " repeated occurrence ", line[1:9]
            data_at_intrinsic[tuple(line[2:])].append(line[:2])
        else:
#            print " new key ", line[1:9]
            data_at_intrinsic[tuple(line[2:])] = [line[:2]]
      except:
          continue

header_str=''
with open(opts.fname[0][0],'r') as f:
    header_str = f.readline()
    header_str = header_str.rstrip()

if header_str.startswith('#'):
    print(header_str)  # repeat the header, so we recapture al the syntax
for key in data_at_intrinsic:
    lnL, sigmaOverL =   np.transpose(data_at_intrinsic[key])
    lnLmax = np.max(lnL)
    sigma = sigmaOverL*np.exp(lnL-lnLmax)  # remove overall Lmax factor, which factors out from the weights constructed from \sigma

    # This is an average, treating them as independent measurements
    wts = weight_simulations.AverageSimulationWeights(None, None,sigma)   
    lnLmeanMinusLmax = scipy.special.logsumexp((lnL - lnLmax)*wts)
    sigmaNetOverL = (np.sqrt(1./np.sum(1./sigma/sigma)))/np.exp(lnLmeanMinusLmax)

    if opts.combination =='sum':
        lnLmeanMinusLmax *= len(wts)   # not an average !
        sigmaNetOverL *= np.sqrt(len(wts))
                              
    print(" {} {} ".format(lnLmeanMinusLmax+lnLmax, sigmaNetOverL) + ' '.join(map(str,key)) )



# weight_simulations.py
#
#  Library to provide black-box, single-point-of-contact tool to weight results of multiple ILE runs (at multiple intrinsic parameters)
#  to produce a composite result.  Provides
#      - weighting by priors over intrinsic parameters
#      - uniform or variance-minimizing 
#      

import numpy as np

def AverageSimulationWeights(param_list, length_list,sigma_list, prior_volume_list=None, weight_at_intrinsic="sigma"):
    """
    AverageSimulationWeights returns the factors W_k appearing in a simple average over
    N simulations, for example for the integral value
                   I = \sum_k W_k I_k
    For backward compatibility, weights can be chosen that depend on
         - the number of draws for each simulation
         - the variance of each simulation
         - the parameter values of each simulation
         - priors,  *if* the priors are provided as *pre-integrated volumes* (i.e., an array,
           one member per simulation).  The user is assumed to have already identified volumes surrounding each simulation and 
           integrated the prior accordingly.  The user is responsible for the (in)accuracy of the resulting average, if the prior and likelihood
           vary significantly versus intrinsic parameters.
    Specific implementations and use cases include:
         - single ILE run at fixed parameters, with different numbers of draws, either by variance-minimizing
                 I = (\sum_k  I_k /\sigma_k^2) / \sum_k 1/\sigma_k^2    
           or by a simple average over the N simulations
                 I = (\sum_k  I_k)/N         N=# of simulations

         - multiple ILE runs at different parameters, accounting for an intrinsic prior volume \Delta P_k for each simulation
           (normalized relative to the intrinsic prior: \sum_k \Delta P_k = (prior probability of lying in ellipsoid) << 1)
                 I =( \sum_k I_k \Delta P_k /\sigma_k^2)/ \sum_k 1/\sigma_k^2
    """

    if np.min(sigma_list) <0.:
        return np.ones(len(sigma_list))   # Allow the user to be a moron

    weights = np.ones(len(sigma_list))

    # Weighting at intrinsic parameters
    if (weight_at_intrinsic=="sigma"):
        weights *= 1./np.power(sigma_list,2)
        weights *= 1./np.sum(weights)          # explicitly renormalize *by intent*
    elif (weight_at_intrinsic=="uniform") or True:
        weights *= 1./len(weights)
    
    
    # Factoring in prior volume.  Note the 'box' average is less *significantly* less accurate than interpolating over prior parameters
    if prior_volume_list:
        weights *= prior_volume_list

    return weights

